/*
 * main.c
 *
 * The first task lets LED4 blink fast. The second task lets LED4 blink slowly.
 * As both tasks use LED4 it is a common resource. Common resources have to be
 * protected by mutexes.
 *
 * Acquired skill: Understand how to protect common resources by mutexes.
 *
 * Please see www.freertos.org for more details.
 * xSemaphoreTake: FreeRTOS/API Reference/Semaphore / Mutexes
 * xSemaphoreGive: FreeRTOS/API Reference/Semaphore / Mutexes
 * xSemaphoreCreateMutex: FreeRTOS/API Reference/Semaphore / Mutexes
 *
 *  Created: 2015-01-29
 * Modified: 2018-07-01
 *   Author: Wolfgang Neff
 */ 

#include <avr/io.h>
#include <FreeRTOS.h>
#include <semphr.h>
#include <task.h>

#include "board.h"

// A mutex is a special kind of semaphore.
SemaphoreHandle_t xMutex;

// The fast blink task checks button 0 and lets blink LED4 25 times
// if the button is pressed.
void vBlinkFastTask(void* pvParameters)
{
	while (1) {
		// Check if button 0 is pressed
		if (!(BUTTON_LOW_PORT.IN & BUTTON0_PIN_bm)) {
			// Ask for the mutex.
			xSemaphoreTake(xMutex, portMAX_DELAY);
			// We have got the mutex now and let the LED blink.
			for (uint8_t i=0; i<50; i++) {
				LED_PORT.OUT ^= LED4_PIN_bm;
				vTaskDelay(pdMS_TO_TICKS(100));
			}
			// Give the mutex back.
			xSemaphoreGive(xMutex);
		}
		// Wait a bit before checking the button again.
		vTaskDelay(pdMS_TO_TICKS(100));
	}
}

// The slow blink task checks button 1 and lets blink LED4 fife times
// if the button is pressed.
void vBlinkSlowTask(void* pvParameters)
{
	while (1) {
		// Check if button 1 is pressed
		if (!(BUTTON_LOW_PORT.IN & BUTTON1_PIN_bm)) {
			// Ask for the mutex.
			xSemaphoreTake(xMutex, portMAX_DELAY);
			// We have got the mutex now and let the LED blink.
			for (uint8_t i=0; i<10; i++) {
				LED_PORT.OUT ^= LED4_PIN_bm;
				vTaskDelay(pdMS_TO_TICKS(500));
			}
			// Give the mutex back.
			xSemaphoreGive(xMutex);
		}
		// Wait a bit before checking the button again.
		vTaskDelay(pdMS_TO_TICKS(100));
	}
}

int main(void)
{
	// Initialize the LED.
    LED_PORT.DIR = LED4_PIN_bm;
    LED_PORT.OUT = LED4_PIN_bm;

	// Initialize the push-buttons.
	BUTTON0_PINCTRL = PORT_OPC_PULLUP_gc;
	BUTTON1_PINCTRL = PORT_OPC_PULLUP_gc;

	// Create the mutex.
	xMutex = xSemaphoreCreateMutex();

	// Create the tasks.
    xTaskCreate(vBlinkFastTask, "vBlinkFastTask", configMINIMAL_STACK_SIZE, NULL, tskIDLE_PRIORITY+1, NULL);
    xTaskCreate(vBlinkSlowTask, "vBlinkSlowTask", configMINIMAL_STACK_SIZE, NULL, tskIDLE_PRIORITY+1, NULL);

	// Start the scheduler.
    vTaskStartScheduler();
}