/*
 * main.c
 *
 * The first task lets LED4 blink fast. The second task lets LED4 blink slowly.
 * The third task checks the buttons. It tells the other tasks whether to blink
 * fast or slowly with the aid of shared memory. Shared memory is a common
 * resource an has therefore to be protected by mutexes.
 *
 * Acquired skill: Understand how to to exchange information between tasks with
 * shared memory.
 *
 * Created: 2018-07-01
 *  Author: Wolfgang Neff
 */ 

#include <avr/io.h>
#include <FreeRTOS.h>
#include <semphr.h>
#include <task.h>

#include "board.h"
#include "switch.h"

// This global memory is the shared memory. We declare it volatile as
// other tasks can change it unexpectedly. We use the following states:
// 0 -> do not blink
// 1 -> blink fast
// 2 -> blink slowly
// Initial state: do not blink.
volatile uint8_t iBlinkState = 0;

// This mutex will protect the shared memory.
SemaphoreHandle_t xMutex;

// The fast blink task checks if it has work and lets then
// blink LED4 25 times. If finished it resets the blink state.
void vBlinkFastTask(void* pvParameters)
{
	while (1) {
		// Check the blink state.
		if (iBlinkState == 1) {
			// Ask for access to the shared memory.
			xSemaphoreTake(xMutex, portMAX_DELAY);
			// Access has been granted. We let the LED blink, now.
			for (uint8_t i=0; i<50; i++) {
				LED_PORT.OUT ^= LED4_PIN_bm;
				vTaskDelay(pdMS_TO_TICKS(100));
			}
			// The job has be done. Blinking has finished.
			iBlinkState = 0;
			// Access to the shared memory is not longer required.
			xSemaphoreGive(xMutex);
		}
		// Wait a bit before checking the state again.
		vTaskDelay(pdMS_TO_TICKS(100));
	}
}

// The slow blink task checks if it has work and lets then
// blink LED4 fife times. If finished it resets the blink state.
void vBlinkSlowTask(void* pvParameters)
{
	while (1) {
		// Check the blink state.
		if (iBlinkState == 2) {
			// Ask for access to the shared memory.
			xSemaphoreTake(xMutex, portMAX_DELAY);
			// Access has been granted. We let the LED blink, now.
			for (uint8_t i=0; i<10; i++) {
				LED_PORT.OUT ^= LED4_PIN_bm;
				vTaskDelay(pdMS_TO_TICKS(500));
			}
			// The job has be done. Blinking has finished.
			iBlinkState = 0;
			// Access to the shared memory is not longer required.
			xSemaphoreGive(xMutex);
		}
		// Wait a bit before checking the state again.
		vTaskDelay(pdMS_TO_TICKS(100));
	}
}

// The button task checks the buttons and changes the state accordingly.
// Button 0 lets the LED blink fast. Button 1 lets the LED blink slowly.
// We use SWITCHn_PRESSED for a better handling of the buttons.
void vButtonTask(void* pvParameters)
{
	while (1) {
		// Check button 0.
		if (SWITCH0_PRESSED()) {
			// Ask for access to the shared memory.
			xSemaphoreTake(xMutex, portMAX_DELAY);
			// Access has been granted. We can change the state, now.
			iBlinkState = 1;
			// Access to the shared memory is not longer required.
			xSemaphoreGive(xMutex);
		}
		// Check button 1.
		if (SWITCH1_PRESSED()) {
			// Ask for access to the shared memory.
			xSemaphoreTake(xMutex, portMAX_DELAY);
			// Access has been granted. We can change the state, now.
			iBlinkState = 2;
			// Access to the shared memory is not longer required.
			xSemaphoreGive(xMutex);
		}
		// Wait a bit before checking the button again.
		vTaskDelay(pdMS_TO_TICKS(25));
	}
}

int main(void)
{
	// Initialize the LED.
    LED_PORT.DIR = LED4_PIN_bm;
    LED_PORT.OUT = LED4_PIN_bm;

	// Initialize the push-buttons.
	BUTTON0_PINCTRL = PORT_OPC_PULLUP_gc;
	BUTTON1_PINCTRL = PORT_OPC_PULLUP_gc;

	// Create the mutex.
	xMutex = xSemaphoreCreateMutex();

	// Create the tasks.
    xTaskCreate(vBlinkFastTask, "vBlinkFastTask", configMINIMAL_STACK_SIZE, NULL, tskIDLE_PRIORITY+2, NULL);
    xTaskCreate(vBlinkSlowTask, "vBlinkSlowTask", configMINIMAL_STACK_SIZE, NULL, tskIDLE_PRIORITY+2, NULL);
    xTaskCreate(vButtonTask, "vButtonTask", configMINIMAL_STACK_SIZE, NULL, tskIDLE_PRIORITY+1, NULL);

	// Start the scheduler.
    vTaskStartScheduler();
}