/*
 * main.c
 *
 * The first task handles LED4. The second task handles LED5. The third task
 * checks the buttons. It tells the other tasks whether to blink fast or slowly.
 * The button task and the LED tasks communicate via a message queue. As two
 * tasks read the queue we do not know which LED will blink.
 *
 * Acquired skill: Understand how to to exchange information between tasks with
 * a message queue.
 *
 * Please see www.freertos.org for more information.
 * xQueueSend: FreeRTOS/API Reference/Queues
 * xQueueReceive: FreeRTOS/API Reference/Queues
 * xQueueCreate: FreeRTOS/API Reference/Queues
 *
 *  Created: 2015-01-27
 * Modified: 2018-07-01
 *   Author: Wolfgang Neff
 */ 

#include <avr/io.h>
#include <FreeRTOS.h>
#include <queue.h>
#include <task.h>

#include "board.h"
#include "switch.h"

// Declare the message queue.
QueueHandle_t xQueue;

// This task reads the message queue and handles LED4.
void vLED4Task(void* pvParameters)
{
    // Message buffer.
    uint8_t message;
	while (1) {
		// Receive message and write it into the message buffer.
		xQueueReceive(xQueue, &message, portMAX_DELAY);
		// Blink according to the message.
		if (message == 1) {
			for (uint8_t i=0; i<50; i++) {
				LED_PORT.OUT ^= LED4_PIN_bm;
				vTaskDelay(pdMS_TO_TICKS(100));
			}
		}
		if (message == 2) {
			for (uint8_t i=0; i<10; i++) {
				LED_PORT.OUT ^= LED4_PIN_bm;
				vTaskDelay(pdMS_TO_TICKS(500));
			}
		}
	}
}

// This task reads the message queue and handles LED5.
void vLED5Task(void* pvParameters)
{
    // Message buffer.
    uint8_t message;
    while (1) {
		// Receive message and write it into the message buffer.
	    xQueueReceive(xQueue, &message, portMAX_DELAY);
		// Blink according to the message.
	    if (message == 1) {
		    for (uint8_t i=0; i<50; i++) {
			    LED_PORT.OUT ^= LED5_PIN_bm;
			    vTaskDelay(pdMS_TO_TICKS(100));
		    }
	    }
	    if (message == 2) {
		    for (uint8_t i=0; i<10; i++) {
			    LED_PORT.OUT ^= LED5_PIN_bm;
			    vTaskDelay(pdMS_TO_TICKS(500));
		    }
	    }
    }
}

// The button task checks the buttons and places a message into the queue.
// Message 1 tells the LED task to blink fast. Message 2 tells it to blink
// slowly. We use SWITCHn_PRESSED for a better handling of the buttons.
void vButtonTask(void* pvParameters)
{
    // Message buffer.
    uint8_t message;
	while (1) {
		// Check button 0.
		if (SWITCH0_PRESSED()) {
            // We have to pass a pointer. Therefore we cannot send
			// the message directly.
			message = 1;
			xQueueSend(xQueue, &message, portMAX_DELAY);
		}
		// Check button 1.
		if (SWITCH1_PRESSED()) {
			message = 2;
			xQueueSend(xQueue, &message, portMAX_DELAY);
		}
		// Wait a bit before checking the button again.
		vTaskDelay(pdMS_TO_TICKS(25));
	}
}

int main(void)
{
	// Initialize the LEDs.
    LED_PORT.DIR = (LED4_PIN_bm | LED5_PIN_bm);
    LED_PORT.OUT = (LED4_PIN_bm | LED5_PIN_bm);

	// Initialize the push-buttons.
	BUTTON0_PINCTRL = PORT_OPC_PULLUP_gc;
	BUTTON1_PINCTRL = PORT_OPC_PULLUP_gc;

	// Create the message queue.
	xQueue = xQueueCreate(4, sizeof(uint8_t));

	// Create the tasks.
    xTaskCreate(vLED4Task, "vLED4Task", configMINIMAL_STACK_SIZE, NULL, tskIDLE_PRIORITY+2, NULL);
    xTaskCreate(vLED5Task, "vLED5Task", configMINIMAL_STACK_SIZE, NULL, tskIDLE_PRIORITY+2, NULL);
    xTaskCreate(vButtonTask, "vButtonTask", configMINIMAL_STACK_SIZE, NULL, tskIDLE_PRIORITY+1, NULL);

	// Start the scheduler.
    vTaskStartScheduler();
}
